#! /usr/bin/env python

import sys, os, subprocess, shutil 
from collections import Counter
import pysam

from . import smith_waterman

def reverse_complement(seq):
    complement = {'A': 'T', 'C': 'G', 'G': 'C', 'T': 'A',
                  'W': 'W', 'S': 'S', 'M': 'K', 'K': 'M', 'R': 'Y', 'Y': 'R',
                  'B': 'V', 'V': 'B', 'D': 'H', 'H': 'D', 'N': 'N'}

    return("".join(complement.get(base, base) for base in reversed(seq)))



def get_consensus_from_mafft_result(input_file):

    id2seq = {}
    with open(input_file, 'r') as hin:
        for line in hin:
            line = line.rstrip('\n')
            if line.startswith('>'):
                tid = line
                id2seq[tid] = ''
            else:
                id2seq[tid] = id2seq[tid] + line


    ind2bases = {}
    for tid in id2seq:
        seq = id2seq[tid]
        for i in range(len(seq)):
            if i not in ind2bases: ind2bases[i] = []
            ind2bases[i].append(seq[i])


    seq_len = len(list(ind2bases))
    consensus = ''
    for i in range(seq_len):
        # import pdb; pdb.set_trace()
        mycounter = Counter(ind2bases[i] )
        consensus = consensus + mycounter.most_common()[0][0]

    consensus = consensus.replace('-', '').upper()

    return(consensus)



def get_refined_bp(contig, fasta_file_ins, chr1, start1, end1, dir1, chr2, start2, end2, dir2, mode, h_log, rd_margin = 20, i_margin = 500):

    start1 = max(1, start1)
    start2 = max(1, start2)

    if mode != "i":

        bstart1, bend1 = max(int(start1) - rd_margin, 1), int(end1) + rd_margin
        bstart2, bend2 = max(int(start2) - rd_margin, 1), int(end2) + rd_margin

        region1_seq = fasta_file_ins.fetch(chr1, bstart1 - 1, bend1)
        region2_seq = fasta_file_ins.fetch(chr2, bstart2 - 1, bend2)

        if dir1 == '-': region1_seq = reverse_complement(region1_seq)
        if dir2 == '+': region2_seq = reverse_complement(region2_seq)

        sret = smith_waterman.sw_jump(contig, region1_seq, region2_seq)
        if sret is None: return(None)
        score, contig_align, region1_align, region2_align, contig_seq, region_seq = sret

        bp_pos1 = bstart1 + region1_align[1] - 1 if dir1 == '+' else bend1 - region1_align[1] + 1 
        bp_pos2 = bstart2 + region2_align[0] - 1 if dir2 == '-' else bend2 - region2_align[0] + 1

        if contig_align[2] - contig_align[1] == 1:
            inseq = '---'
        elif contig_align[2] - contig_align[1] > 1:
            inseq = contig[(contig_align[1]):(contig_align[2] - 1)]
            if dir1 == '-': inseq = reverse_complement(inseq)
        else:
            print("Alignment inconsistent!!", file = sys.stderr)

        print(score, contig_align, region1_align, region2_align, file = h_log)
        print(contig_seq, file = h_log)
        print(region_seq, file = h_log)

        return(bp_pos1, bp_pos2, inseq)
    
    else:
    
        contig_start = contig[:min(i_margin, len(contig))]
        contig_end = contig[-min(i_margin, len(contig)):]

        region_seq = fasta_file_ins.fetch(chr1, max(0, start1 - 100), end2 + 100)
        sret = smith_waterman.sw_jump(region_seq, contig_start, contig_end)
        if sret is None: return(None)
        score, region_align, contig_start_align, contig_end_align, region_seq, contig_seq = sret

        bp_pos1 = start1 - 100 + region_align[1]
        bp_pos2 = start1 - 100 + region_align[2]

        inseq_start = contig_start_align[1]
        inseq_end = len(contig) - (len(contig_end) - contig_end_align[0] + 1)
        inseq = contig[inseq_start:(inseq_end + 1)]

        print(score, region_align, contig_start_align, contig_end_align, file = h_log)
        print(region_seq, file = h_log)
        print(contig_seq, file = h_log)

        return(bp_pos1, bp_pos2, inseq)

        
def get_readid2alignment(input_file, mode, alignment_margin):

    readid2alignment = {}
    with open(input_file, 'r') as hin:
        for line in hin:
            F = line.rstrip('\n').split('\t')
            key = ','.join([F[0], F[1], F[2], F[8], F[3], F[4], F[5], F[9], mode])
            readids = F[6].split(';')

            if mode == "r":
                info1 = F[10].split(';')
                info2 = F[11].split(';')
                for i in range(len(readids)):
                    tinfo1 = info1[i].split(',') 
                    tinfo2 = info2[i].split(',')
                    start1, end1, start2, end2 = int(tinfo1[0]), int(tinfo1[2]), int(tinfo2[0]), int(tinfo2[2])
                    if readids[i] not in readid2alignment: readid2alignment[readids[i]] = []
                    if start1 <= start2:
                        readid2alignment[readids[i]].append((key, max(end1 - alignment_margin, 0), start2 + alignment_margin, '+'))
                    else:
                        readid2alignment[readids[i]].append((key, max(end2 - alignment_margin, 0), start1 + alignment_margin, '-'))

            elif mode == "d":
                size = F[10].split(';')
                info = F[11].split(';')
                for i in range(len(readids)):
                    tinfo = info[i].split(',')
                    tpos, tlen, tstrand = int(tinfo[1]), int(tinfo[3]), tinfo[4]
                    if readids[i] not in readid2alignment: readid2alignment[readids[i]] = []
                    readid2alignment[readids[i]].append((key, max(tpos - alignment_margin, 0), min(tpos + alignment_margin, tlen - 1), tstrand))

            elif mode == "i":
                size = F[10].split(';')
                info = F[11].split(';')
                for i in range(len(readids)):
                    tinfo = info[i].split(',')
                    tpos, tlen, tstrand = int(tinfo[1]), int(tinfo[3]), tinfo[4]
                    if readids[i] not in readid2alignment: readid2alignment[readids[i]] = []
                    if tstrand == "+":
                        readid2alignment[readids[i]].append((key, max(tpos - alignment_margin, 0), min(tpos + int(size[i]) + alignment_margin, tlen - 1), tstrand))
                    else:
                        readid2alignment[readids[i]].append((key, max(tpos - int(size[i]) - alignment_margin, 0), min(tpos + alignment_margin, tlen - 1), tstrand))

    return(readid2alignment)

 
def identify(rearrangement_file, insertion_file, deletion_file, output_file, tumor_bam, reference_fasta, debug, alignment_margin = 300):

    bamfile = pysam.AlignmentFile(tumor_bam, "rb")

    readid2alignment = get_readid2alignment(rearrangement_file, 'r', alignment_margin)
    readid2alignment.update(get_readid2alignment(insertion_file, 'i', alignment_margin))
    readid2alignment.update(get_readid2alignment(deletion_file, 'd', alignment_margin))

    hout = open(output_file + ".tmp.supporting_read.unsorted", 'w')
    for read in bamfile.fetch():

        if read.query_name in readid2alignment and not read.is_secondary and not read.is_supplementary:

            for talignment in readid2alignment[read.query_name]:

                tkey, tstart, tend, tstrand = talignment

                read_seq = reverse_complement(read.query_sequence) if read.is_reverse else read.query_sequence
                part_seq = read_seq[(tstart - 1):tend]
                if tstrand == '-': part_seq = reverse_complement(part_seq)
                # if read.is_reverse: part_seq = reverse_complement(part_seq)

                print('\t'.join([tkey, read.query_name, part_seq]), file = hout)

    hout.close()

    hout = open(output_file + ".tmp.supporting_read.sorted", 'w')
    subprocess.check_call(["sort", "-k1,1", output_file + ".tmp.supporting_read.unsorted"], stdout = hout)
    hout.close()
    subprocess.check_call(["rm", "-rf", output_file + ".tmp.supporting_read.unsorted"])
    
    fasta_file_ins = pysam.FastaFile(reference_fasta)

    tmp_dir = output_file + ".tmp_alignment_dir"
    if os.path.exists(tmp_dir): shutil.rmtree(tmp_dir)
    os.makedirs(tmp_dir)

    temp_key = ''
    # hout = open(output_file + ".tmp.consensus.fastq", 'w') 
    hout_log = open(tmp_dir + "/consensus_alignment.log", 'w')

    hout = open(output_file, 'w') 
    with open(output_file + ".tmp.supporting_read.sorted", 'r') as hin:
        for line in hin:
            F = line.rstrip('\n').split('\t')
            if temp_key != F[0]: 

                if temp_key != '':
                    hout2.close()

                    hout3 = open(tmp_dir + '/' + temp_key + ".mafft_result.fa", 'w')
                    subprocess.check_call(["mafft", tmp_dir + '/' + temp_key + ".supporting_read.fa"], stdout = hout3, stderr = subprocess.DEVNULL)
                    hout3.close()

                    tconsensus = get_consensus_from_mafft_result(tmp_dir + '/' + temp_key + ".mafft_result.fa")
                    print(temp_key + '\n' + tconsensus, file = hout_log)
                    # print(temp_key + '\n' + tconsensus)
         
                    chr1, start1, end1, dir1, chr2, start2, end2, dir2, mode = temp_key.split(',')
                    start1, end1, start2, end2 = int(start1), int(end1), int(start2), int(end2)       
                    bret = get_refined_bp(tconsensus, fasta_file_ins, chr1, start1, end1, dir1, chr2, start2, end2, dir2, mode, hout_log)
                    if bret is not None:  
                        bp_pos1, bp_pos2, inseq = bret 
                        print(bp_pos1, bp_pos2, inseq, file = hout_log)
                        print('', file = hout_log)
                        print('\t'.join([chr1, str(bp_pos1), dir1, chr2, str(bp_pos2), dir2, inseq]), file = hout)

                temp_key = F[0]          
                hout2 = open(tmp_dir + '/' + temp_key + ".supporting_read.fa", 'w')

            print('>' + F[1] + '\n' + F[2], file = hout2)

        # last treatment
        if temp_key != '':
            hout2.close()

            hout3 = open(tmp_dir + '/' + temp_key + ".mafft_result.fa", 'w')
            subprocess.check_call(["mafft", tmp_dir + '/' + temp_key + ".supporting_read.fa"], stdout = hout3, stderr = subprocess.DEVNULL)
            hout3.close()

            tconsensus = get_consensus_from_mafft_result(tmp_dir + '/' + temp_key + ".mafft_result.fa")
            print(temp_key + '\n' + tconsensus, file = hout_log)

            chr1, start1, end1, dir1, chr2, start2, end2, dir2, mode = temp_key.split(',')     
            start1, end1, start2, end2 = int(start1), int(end1), int(start2), int(end2)
            bret = get_refined_bp(tconsensus, fasta_file_ins, chr1, start1, end1, dir1, chr2, start2, end2, dir2, mode, hout_log)
            if bret is not None:
                bp_pos1, bp_pos2, inseq = bret
                print(bp_pos1, bp_pos2, inseq, file = hout_log)
                print('', file = hout_log)
                print('\t'.join([chr1, str(bp_pos1), dir1, chr2, str(bp_pos2), dir2, inseq]), file = hout)

    hout_log.close()
    hout.close()

    if not debug:
        subprocess.check_call(["rm", "-rf", output_file + ".tmp.supporting_read.sorted"])
        shutil.rmtree(tmp_dir)


if __name__ == "__main__":

    """
    import sys
    rearrangement_file = sys.argv[1]
    deletion_file = sys.argv[2]
    output_file = sys.argv[3]
    tumor_bam = sys.argv[4]
    reference_fasta = sys.argv[5]

    identify(rearrangement_file, deletion_file, output_file, tumor_bam, reference_fasta)

    key = "X,104681514,104681542,+,X,104681515,104681543,-,i"
    tconsensus = "ATATGATTTAATAACCAATTCACAGAAAAGTTTAAATGACTGATAATAGCTACATGTGTTTCCTCTAGTAGGGCCATATTCATTTGGGTAGGTTACCAAAATCTTTAAATTAAAATAACGCATTTTCAACAGTAGGTGGGATGCCAAGCAACAACCAAAGGATAAAGGAGTTGGTAGGGATAAGAGTCAGAACAGCCTCCATTACATTACTGAAATAGAAGATATATTTGCTTTAAAAATTTTAATTTGGAGGTTGTATAAGAATATCACTTAGTCTATATGGTACTATAGACAAATATATATAGAAATGTAGGCATTTAGTGCTATAAATTTCCCTCTACACACTGCTTTGAATGGGTCCCAGAGATTCTGGTATGTGGTGTCTTTGTTCTCGTTGGTTTCAAAGAACATCTTTATTTCTGCCTTCATTTCAATATGTACCCAGTAGTCATTCAGGAGCAGGTTGTTCAGTTTCCAGTGTAGTTGAGCAGCTTTGGTGAGATTCTTAATCTGAGTTCTAGTTTGATTGCACTGTGGTCTGAGAGATAGTTTGTTATAATTTCTGTTCTTTTACATTTGCTGAGGAGATTTACTTCAACTATGTGGTCAATTTTGGAATAGGTGTGGTGTGGTGCTGAAAAAAATGTATATTCTGTTGATTTGGGGGAGAGTTCTGTAGATGTCTATTAGGTCTGCTTGGTGAGGGTTCAATCCTGGGTATCCTTGACTGGATTAAGAAAATGTGGCTACACCATGGAATACTATGCAGCCATAAAATGATGAGTTCATATCCTTTAAGGGACAATGAATGAAACATCATTCTCAGTAAACTATCATAAAACAAAAACAAAGCATATTCTACTCATAGGTGGGAATTGAACAATGAGTCACATGGACACAGGAGGAATATCACCTCTGGGGACTGTGGTGGGGTCGGGGGGGAGGATAGCATTGGGAATATACCTAATTAGATGACACATAGTGGGTGCAGCGCCAGCATGCACATGTATACATGTAACTAACCTGCCAATGTGCATGTGTACTAAAACTTAGGTATAATAAAAAAAAAAAAAAATAATAATAATAAAAAAAAAAAAAAAAAAAAAAAAAAAATATAGTAACACTCCCAAATACTGATGAATGTGCTCAGACCAGGAGAAATCTTACATGGAAAATAAAATAAGTTCTTGAAATCTTAACTTTAATAAATCATTTAGTTGATTGATTTTACATCAAAATTCTGAAACTAAACTCATTGTAAAATATAAGCGATATATATATATATATACCATATTTTGTCACTTTCTGAGAGCTGAGCAACTATAGTTGCATATAAATCACTTTAAAACTTATGTAATAGTGAGAATTCATTTTAAATTATAGTATTTTTA"
    """

    key = "9,12910928,12910950,+,9,12910929,12910951,-,i"
    tconsensus = "AACACATGAAGTATTTGGATCAGCTGGGTACAGTGGCTCACGCCTCTAATCTCAACACTTTGGGAGGCTGAGGCGGGTGGATCACTGAGGTCAGGAGTTTGAGACCAGCCTGGCCAACATAGTGAAGCTGTTTCTACTAAAATACAAAAATTAGCCAGGCGTGGCGGGCACCTGTAATCCCAGCTACTCAGGAGGCTAAGGCAGGAGAATCACTTGAACCTGGGAGGCAGAGGTTGCAGTGAGCCGAGATCGTGTCACTACACACTTCAACTTGGGTGACAGACTGAGATATACCTAATGCTAGATGACACATTAGTGGGTGCAGCGCACCAGCATAGCATGTATACATATGTAACTAACCTGCACAATGTGCACATGTCCCTAAAACTTAAGTATAATAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAGTATTTGAATCAAGGTCTTTTGAAGTCTCAGGCATTGACCATCTTGGTCTGTGAAGTCCTCACTATGATAAATATGCCAATTATGCAGCCTTCAAGTAGATCCTCAGTTACATGTACTGAGGCAGGAAAGGACTAAAGCCAAAAGCTCAATTACAGGCACCACTATGGGAGTCAAGCCCCTTATTTGAGGCTGAAACAAAATATCAGAAACCCCCAAGGATAATCCAGACTTCGAAATTAGCAGGCGATCAACTGAAAAGTCAGACATTGTTCTATAT"

    chr1, start1, end1, dir1, chr2, start2, end2, dir2, mode = key.split(',')
    start1, end1, start2, end2 = int(start1), int(end1), int(start2), int(end2) 
    hout_log = open("test.txt", 'w')
    fasta_file_ins = pysam.FastaFile("/home/ubuntu/environment/workspace/seq_data/reference/GRCh37.fa")

    bret = get_refined_bp(tconsensus, fasta_file_ins, chr1, start1, end1, dir1, chr2, start2, end2, dir2, mode, hout_log)
    print(bret)




#! /usr/bin/env python

import argparse

from .run import *
from .version import __version__

def create_parser():

    parser = argparse.ArgumentParser(prog = "nanomonsv")

    parser.add_argument("--version", action = "version", version = "%(prog)s " + __version__)

    subparsers = parser.add_subparsers()

    ##########
    # parse
    parse = subparsers.add_parser("parse",
                                help = "Parse supporting reads for candidate structural variations")


    parse.add_argument("bam_file", default = None, type = str,
                     help = "Path to input BAM file")

    parse.add_argument("output_prefix", type = str,
                       help = "Prefix of output files")

    parse.add_argument("--debug", default = False, action = 'store_true', help = "keep intermediate files")

    parse.add_argument("--split_alignment_check_margin", default = 50, type = int,
                       help = "Two split alignments whose margin sizes are no more than this value is counted as candidate breakpoint")

    parse.add_argument("--minimum_breakpoint_ambiguity", default = 20, type = int,
                       help = "Sizes of ambiguities of breakpoint positions from the observed ones") 

    parse.set_defaults(func = parse_main)
    ##########
    
    ##########
    # get
    get = subparsers.add_parser("get",
                                help = "List up reliable structural variations with refined breakpoint positions")

    get.add_argument("tumor_prefix", type = str,
                      help = "Prefix of tumor data processed in parse step")
       
    get.add_argument("tumor_bam", default = None, type = str,
                      help = "Path to tumor BAM file")
 
    get.add_argument("reference_fasta", metavar = "reference.fa", default = None, type = str,
                     help = "the path to the reference genome sequence")

    get.add_argument("--control_prefix", type = str, # required = True,
                     help = "Prefix of matched control data processed in parse step")

    get.add_argument("--control_bam", type = str, # required = True,
                     help = "Path to control BAM file")

    get.add_argument("--min_tumor_variant_read_num", default = 3, type = int,
                     help = "Minimum required supporting read number for a tumor sample")

    get.add_argument("--min_tumor_VAF", default = 0.05, type = float,
                     help = "Minimum required variant allele frequency for a tumor sample")

    get.add_argument("--max_control_variant_read_num", default = 1, type = int,
                     help = "Maximum allowed supporting read number for a control sample")
        
    get.add_argument("--max_control_VAF", default = 0.03, type = float,
                     help = "Maximum allowed variant allele frequeycy for a control sample")

    get.add_argument("--cluster_margin_size", default = 100, type = int,
                     help = "Two breakpoints are margined if they are within this threshould value")

    # get.add_argument("--read_num_thres", default = 3, type = int,
    #                  help = "Minimum required supporting read number for structural variation candidates")

    get.add_argument("--median_mapQ_thres", default = 20, type = int,
                     help = "Threshould for median mapping quality")

    get.add_argument("--max_overhang_size_thres", default = 100, type = int,
                     help = "Threshould for maximum overhang size")

    get.add_argument("--var_read_min_mapq", default = 0, type = int,
                     help = "Threshould for mapping quality in validate step")

    get.add_argument("--use_ssw_lib", default = False, action = 'store_true',
                     help = "Use SSW Library. This is for backward comaptibility, and may be removed in the future")

    # get.add_argument("--control_read_num_thres", default = 0, type = int,
    #                  help = "Filter if the number of supporting reads for the control sample is larger than this value")

    get.add_argument("--debug", default = False, action = 'store_true', help = "keep intermediate files")

    get.set_defaults(func = get_main)
    ##########

    ##########
    # validate 
    validate = subparsers.add_parser("validate",
                                     help = "Validate GenomonSV format SV list using nanopore sequence reads")
 
    validate.add_argument("sv_list_file", default = None, type = str,
                          help = "Path to GenomonSV format SV list file")

    validate.add_argument("tumor_bam", default = None, type = str,
                          help = "Path to tumor BAM file")

    validate.add_argument("output", type = str,
                       help = "Path to output file")

    validate.add_argument("reference_fasta", metavar = "reference.fa", default = None, type = str,
                          help = "Path to the reference genome sequence")

    validate.add_argument("--control_bam", default = None, type = str,
                          help = "Path to control BAM file")

    validate.add_argument("--var_read_min_mapq", default = 40, type = int,
                          help = "Threshould for mapping quality in validate step")

    validate.add_argument("--use_ssw_lib", default = False, action = 'store_true',
                          help = "Use SSW Library. This is for backward comaptibility, and may be removed in the future")

    validate.add_argument("--debug", default = False, action = 'store_true', help = "keep intermediate files")

    validate.set_defaults(func = validate_main)
    ##########

    ##########
    # insert_classify
    insert_classify = subparsers.add_parser("insert_classify",
                                            help = "Classify long insertion into LINE1, Alu, SVA, and so on")
    
    insert_classify.add_argument("sv_list_file", default = None, type =str,
                                 help = "Path to nanomonsv get result file")
    
    insert_classify.add_argument("output_file", type = str,
                                 help = "Path to output file")

    insert_classify.add_argument("reference_fasta", metavar = "reference.fa", default = None, type = str,
                                 help = "Path to the reference genome sequence")

    insert_classify.add_argument("--grc", default = False, action = 'store_true',
                                 help = "Deprecated. This is not used any more. Convert chromosome names to Genome Reference Consortium nomenclature (default: %(default)s)")

    insert_classify.add_argument("--genome_id", choices = ["hg19", "hg38", "mm10"], default = "hg38",
                                 help = "Genome id used for selecting UCSC-GRC chromosome name corresponding files (default: %(default)s)")

    insert_classify.add_argument("--debug", default = False, action = 'store_true', help = "keep intermediate files")

    insert_classify.set_defaults(func = insert_classify_main)
    ##########

    return parser
  


    



#! /usr/bin/env python

import os, subprocess, shutil
from .parse import *
from .filt import *
from .identify import *
from .long_read_validate import *
from .insert_classify import *
from .utils import *

from .logger import get_logger
logger = get_logger(__name__)

def parse_main(args):

    # check if the executables exist
    is_tool("tabix")
    is_tool("bgzip")

    # check input file existences
    is_exists_bam(args.bam_file)

    # BAM format check
    bam_format_check(args.bam_file)

    # make directory for the output prefix
    output_dir = os.path.dirname(args.output_prefix)
    if output_dir != '' and not os.path.exists(output_dir):
        os.makedirs(output_dir)

    ####################
    parse_alignment_info(args.bam_file, args.output_prefix + ".tmp.deletion_info.txt", 
                                        args.output_prefix + ".tmp.insertion_info.txt", 
                                        args.output_prefix + ".tmp.rearrangement_info.txt")

    ####################
    # deletion processing
    hout = open(args.output_prefix + ".tmp.deletion.sorted.bed", 'w')
    subprocess.check_call(["sort", "-k1,1", "-k2,2n", "-k3,3n", args.output_prefix + ".tmp.deletion_info.txt"], stdout = hout)
    hout.close()

    hout = open(args.output_prefix + ".deletion.sorted.bed.gz", 'w')
    subprocess.check_call(["bgzip", "-f", "-c", args.output_prefix + ".tmp.deletion.sorted.bed"], stdout = hout)
    hout.close()

    subprocess.check_call(["tabix", "-p", "bed", args.output_prefix + ".deletion.sorted.bed.gz"])
    ####################

    ####################
    # insertion processing
    hout = open(args.output_prefix + ".tmp.insertion.sorted.bed", 'w')
    subprocess.check_call(["sort", "-k1,1", "-k2,2n", "-k3,3n", args.output_prefix + ".tmp.insertion_info.txt"], stdout = hout)
    hout.close()
    
    hout = open(args.output_prefix + ".insertion.sorted.bed.gz", 'w') 
    subprocess.check_call(["bgzip", "-f", "-c", args.output_prefix + ".tmp.insertion.sorted.bed"], stdout = hout)
    hout.close()
     
    subprocess.check_call(["tabix", "-p", "bed", args.output_prefix + ".insertion.sorted.bed.gz"])
    ####################

    ####################
    # rearrangement processing
    hout = open(args.output_prefix + ".tmp.rearrangement_info.name_sorted.txt", 'w')
    subprocess.check_call(["sort", "-k1,1", "-k2,2n", args.output_prefix + ".tmp.rearrangement_info.txt"], stdout = hout)
    hout.close()

    extract_bedpe_junction(args.output_prefix + ".tmp.rearrangement_info.name_sorted.txt", 
                           args.output_prefix + ".tmp.rearrangement.bedpe") # ,
                           # args.split_alignment_check_margin, args.minimum_breakpoint_ambiguity)

    hout = open(args.output_prefix + ".tmp.rearrangement.sorted.bedpe", 'w')
    subprocess.check_call(["sort", "-k1,1", "-k2,2n", "-k3,3n", "-k4,4", "-k5,5n", "-k6,6n", 
                           args.output_prefix + ".tmp.rearrangement.bedpe"], stdout = hout)
    hout.close()
  
    hout = open(args.output_prefix + ".rearrangement.sorted.bedpe.gz", 'w')
    subprocess.check_call(["bgzip", "-f", "-c", args.output_prefix + ".tmp.rearrangement.sorted.bedpe"], stdout = hout)
    hout.close()

    subprocess.check_call(["tabix", "-p", "bed", args.output_prefix + ".rearrangement.sorted.bedpe.gz"])
    ####################

    if not args.debug:
        subprocess.check_call(["rm", "-rf", args.output_prefix + ".tmp.deletion_info.txt"])
        subprocess.check_call(["rm", "-rf", args.output_prefix + ".tmp.deletion.sorted.bed"])
        subprocess.check_call(["rm", "-rf", args.output_prefix + ".tmp.insertion_info.txt"])
        subprocess.check_call(["rm", "-rf", args.output_prefix + ".tmp.insertion.sorted.bed"])
        subprocess.check_call(["rm", "-rf", args.output_prefix + ".tmp.rearrangement_info.txt"])
        subprocess.check_call(["rm", "-rf", args.output_prefix + ".tmp.rearrangement_info.name_sorted.txt"])
        subprocess.check_call(["rm", "-rf", args.output_prefix + ".tmp.rearrangement.bedpe"])
        subprocess.check_call(["rm", "-rf", args.output_prefix + ".tmp.rearrangement.sorted.bedpe"])


def get_main(args):

    # check if the executables exist
    is_tool("mafft")
    if args.use_ssw_lib: libssw_check()

    # check existences
    is_exists_bam(args.tumor_bam)
    is_exists(args.reference_fasta)
    if args.control_bam is not None: is_exists_bam(args.control_bam)
   
    # check parsed files existences
    is_exists_parsed_files(args.tumor_prefix)
    if args.control_prefix is not None: is_exists_parsed_files(args.control_prefix)
 
    # BAM format check
    bam_format_check(args.tumor_bam)
    fasta_format_check(args.reference_fasta)
    if args.control_bam is not None: bam_format_check(args.control_bam)

    control_rearrangement_bedpe = None
    control_deletion_bed = None
    control_insertion_bed = None
    if args.control_prefix is not None: 
        control_rearrangement_bedpe = args.control_prefix + ".rearrangement.sorted.bedpe.gz"
        control_deletion_bed = args.control_prefix + ".deletion.sorted.bed.gz"
        control_insertion_bed = args.control_prefix + ".insertion.sorted.bed.gz"
    
    ####################
    cluster_rearrangement(args.tumor_prefix + ".rearrangement.sorted.bedpe.gz", args.tumor_prefix + ".rearrangement.sorted.clustered.bedpe",
                     args.cluster_margin_size)

    filt_clustered_rearrangement1(args.tumor_prefix + ".rearrangement.sorted.clustered.bedpe", args.tumor_prefix + ".rearrangement.sorted.clustered.filt1.bedpe",
                             args.min_tumor_variant_read_num, args.median_mapQ_thres, args.max_overhang_size_thres)

    filt_clustered_rearrangement2(args.tumor_prefix + ".rearrangement.sorted.clustered.filt1.bedpe", args.tumor_prefix + ".rearrangement.sorted.clustered.filt2.bedpe", 
                                  control_rearrangement_bedpe)


    cluster_insertion_deletion(args.tumor_prefix + ".deletion.sorted.bed.gz", args.tumor_prefix + ".deletion.sorted.clustered.bedpe")

    filt_clustered_insertion_deletion1(args.tumor_prefix + ".deletion.sorted.clustered.bedpe", args.tumor_prefix + ".deletion.sorted.clustered.filt1.bedpe",
                                       args.min_tumor_variant_read_num, args.median_mapQ_thres, args.max_overhang_size_thres)

    filt_clustered_insertion_deletion2(args.tumor_prefix + ".deletion.sorted.clustered.filt1.bedpe", args.tumor_prefix + ".deletion.sorted.clustered.filt2.bedpe",
                                       control_deletion_bed)

    cluster_insertion_deletion(args.tumor_prefix + ".insertion.sorted.bed.gz", args.tumor_prefix + ".insertion.sorted.clustered.bedpe")

    filt_clustered_insertion_deletion1(args.tumor_prefix + ".insertion.sorted.clustered.bedpe", args.tumor_prefix + ".insertion.sorted.clustered.filt1.bedpe",
                                       args.min_tumor_variant_read_num, args.median_mapQ_thres, args.max_overhang_size_thres)

    filt_clustered_insertion_deletion2(args.tumor_prefix + ".insertion.sorted.clustered.filt1.bedpe", args.tumor_prefix + ".insertion.sorted.clustered.filt2.bedpe",
                                       control_insertion_bed)

    identify(args.tumor_prefix + ".rearrangement.sorted.clustered.filt2.bedpe", 
             args.tumor_prefix + ".insertion.sorted.clustered.filt2.bedpe",
             args.tumor_prefix + ".deletion.sorted.clustered.filt2.bedpe",
             args.tumor_prefix + ".refined_bp.txt", args.tumor_bam, args.reference_fasta, args.debug)

    long_read_validate_main(args.tumor_prefix + ".refined_bp.txt",
                  args.tumor_bam,
                  args.tumor_prefix + ".refined_bp.validated.txt",
                  args.tumor_prefix + ".validated.tumor_sread.txt",
                  args.reference_fasta, 
                  args.control_bam, args.var_read_min_mapq, args.use_ssw_lib, args.debug)

    is_control = True if args.control_bam is not None else False

    filt_final(args.tumor_prefix + ".refined_bp.validated.txt",
               args.tumor_prefix + ".validated.tumor_sread.txt",
               args.tumor_prefix + ".nanomonsv.result.txt",
               args.tumor_prefix + ".nanomonsv.supporting_read.txt",
               args.min_tumor_variant_read_num, args.min_tumor_VAF, args.max_control_variant_read_num, args.max_control_VAF, False, is_control)

    if not args.debug:
        subprocess.check_call(["rm", "-rf", args.tumor_prefix + ".rearrangement.sorted.clustered.bedpe"])
        subprocess.check_call(["rm", "-rf", args.tumor_prefix + ".rearrangement.sorted.clustered.filt1.bedpe"])
        subprocess.check_call(["rm", "-rf", args.tumor_prefix + ".rearrangement.sorted.clustered.filt2.bedpe"])
        subprocess.check_call(["rm", "-rf", args.tumor_prefix + ".deletion.sorted.clustered.bedpe"])
        subprocess.check_call(["rm", "-rf", args.tumor_prefix + ".deletion.sorted.clustered.filt1.bedpe"])
        subprocess.check_call(["rm", "-rf", args.tumor_prefix + ".deletion.sorted.clustered.filt2.bedpe"])
        subprocess.check_call(["rm", "-rf", args.tumor_prefix + ".insertion.sorted.clustered.bedpe"])
        subprocess.check_call(["rm", "-rf", args.tumor_prefix + ".insertion.sorted.clustered.filt1.bedpe"])
        subprocess.check_call(["rm", "-rf", args.tumor_prefix + ".insertion.sorted.clustered.filt2.bedpe"])
        subprocess.check_call(["rm", "-rf", args.tumor_prefix + ".refined_bp.txt"])
        subprocess.check_call(["rm", "-rf", args.tumor_prefix + ".refined_bp.validated.txt"])
        subprocess.check_call(["rm", "-rf", args.tumor_prefix + ".validated.tumor_sread.txt"])


def validate_main(args):
    
    # executable check
    if args.use_ssw_lib: libssw_check()

    ####################
    long_read_validate_main(args.sv_list_file,
                            args.tumor_bam,
                            args.output + ".validated.txt",
                            args.output + ".validated.tumor_sread.txt",
                            args.reference_fasta,
                            args.control_bam, 
                            args.var_read_min_mapq,
                            args.use_ssw_lib, args.debug)

    is_control = True if args.control_bam is not None else False

    filt_final(args.output + ".validated.txt",
               args.output + ".validated.tumor_sread.txt",
               args.output, 
               args.output + ".supporting_read.txt",
               0, 0, float("inf"), float("inf"), True, is_control)

    if not args.debug:
        subprocess.check_call(["rm", "-rf", args.output + ".validated.txt"])
        subprocess.check_call(["rm", "-rf", args.output + ".validated.tumor_sread.txt"])


def insert_classify_main(args):

    import tempfile
    import annot_utils.exon

    # check if the executables exist
    is_tool("minimap2")
    is_tool("bedtools")
    is_tool("bwa")
    is_tool("RepeatMasker")

    make_fasta_file(args.sv_list_file, args.output_file + ".tmp.fasta", args.output_file + ".tmp.seq_id.txt")
    
    ##########
    # processed pseudo gene
    annot_utils.exon.make_exon_info(args.output_file + ".tmp.exon.bed.gz", "gencode", args.genome_id, args.grc, True)

    with open(args.output_file + ".tmp.minimap2.sam", 'w') as hout:
        subprocess.check_call(["minimap2", "-ax", "splice", args.reference_fasta, args.output_file + ".tmp.fasta"], stdout = hout)

    sam2bed_split(args.output_file + ".tmp.minimap2.sam", args.output_file + ".tmp.minimap2.filt.bed")
    
    with open(args.output_file + ".tmp.minimap2.filt.exon.bed", 'w') as hout:
        subprocess.check_call(["bedtools", "intersect", "-a", args.output_file + ".tmp.minimap2.filt.bed", 
                               "-b", args.output_file + ".tmp.exon.bed.gz", "-wo"], stdout = hout)
 
    pp_proc_filt_exon(args.output_file + ".tmp.minimap2.filt.exon.bed", 
                      args.output_file + ".tmp.seq_id.txt", 
                      args.output_file + ".tmp.ppseudo.txt")
    ##########

    ##########
    # repeat masker
    output_dir = os.path.dirname(args.output_file)
    tmpdir_rmsk = tempfile.mkdtemp()
    # tmpdir_rmsk = "tmp"
    subprocess.check_call(["RepeatMasker", "-species", "human", args.output_file + ".tmp.fasta", "-dir", tmpdir_rmsk])

    summarize_rmsk(tmpdir_rmsk + '/' + os.path.basename(args.output_file + ".tmp.fasta") + ".out", args.output_file + ".tmp.rmsk.txt")

    check_tsd_polyAT(args.output_file + ".tmp.fasta", args.output_file + ".tmp.seq_id.txt", 
                     args.reference_fasta, args.output_file + ".tmp.tsd.polyAT.txt")

    shutil.rmtree(tmpdir_rmsk)
    ##########
    
    ##########
    # alignment to reference genome
    with open(args.output_file + ".tmp.bwa.sam", 'w') as hout:
        print(' '.join(["bwa", "mem", "-h", "200", args.reference_fasta, args.output_file + ".tmp.fasta"]))
        subprocess.check_call(["bwa", "mem", "-h", "200", args.reference_fasta, args.output_file + ".tmp.fasta"], stdout = hout)

    summarize_bwa_alignment2(args.output_file + ".tmp.bwa.sam", args.output_file + ".tmp.seq_id.txt", args.output_file + ".tmp.alignment.txt")

    organize_info(args.output_file + ".tmp.rmsk.txt", args.output_file + ".tmp.alignment.txt", 
                 args.output_file + ".tmp.tsd.polyAT.txt", args.output_file + ".tmp.seq_id.txt", 
                 args.output_file + ".tmp.org.txt", args.genome_id)

    annotate_sv_file(args.sv_list_file, args.output_file + ".tmp.org.txt", args.output_file + ".tmp.ppseudo.txt",
                     args.output_file + ".tmp.seq_id.txt", args.output_file)

    if not args.debug:
        os.remove(args.output_file + ".tmp.fasta")
        os.remove(args.output_file + ".tmp.seq_id.txt")
        os.remove(args.output_file + ".tmp.bwa.sam")
        os.remove(args.output_file + ".tmp.exon.bed.gz")
        os.remove(args.output_file + ".tmp.exon.bed.gz.tbi")
        os.remove(args.output_file + ".tmp.minimap2.filt.bed")
        os.remove(args.output_file + ".tmp.minimap2.filt.exon.bed")
        os.remove(args.output_file + ".tmp.ppseudo.txt")
        os.remove(args.output_file + ".tmp.rmsk.txt")
        os.remove(args.output_file + ".tmp.tsd.polyAT.txt")
        os.remove(args.output_file + ".tmp.minimap2.sam")
        os.remove(args.output_file + ".tmp.alignment.txt")
        os.remove(args.output_file + ".tmp.org.txt")


